from time import sleep

import click

from orc_dev_cli import __version__
from orc_dev_cli.code import (
    cli_addon,
    cli_cluster_state,
    cli_creds,
    cli_delete,
    state_exit_condition,
)
from orc_dev_cli.config import configuration_file, load_config, try_function_defined
from orc_dev_cli.index_build.index import cli_index, cli_template

CONFIG = load_config()


def abort_if_false(ctx, param, value):
    if not value:
        ctx.abort()


@click.group()
@click.version_option(version=__version__)
def cli():
    """
    Dev tooling for creating, loging into, deleting OSD cluster and watching RHOAM addons

    Configuration file location: ~/.config/orc/config.toml
    """
    pass


@cli.command()
@click.option(
    "-c",
    "--cluster",
    default=try_function_defined(CONFIG, "login", "cluster"),
    show_default=True,
    help="Name of cluster. Config: cluster",
)
def login(cluster):
    """
    Get cluster kubeadmin login details.
    """
    cli_creds(cluster)


@cli.command()
@click.option(
    "-c",
    "--cluster",
    default=try_function_defined(CONFIG, "osd", "cluster"),
    show_default=True,
    help="Name of cluster. Config: cluster",
)
@click.option(
    "-w/-x",
    "--watch/--no-watch",
    default=try_function_defined(CONFIG, "osd", "watch"),
    help="Watch the current state of a cluster, Config: watch [boolean]",
    show_default=True,
)
@click.option(
    "-d",
    "--delay",
    type=int,
    default=try_function_defined(CONFIG, "osd", "delay"),
    help="Refresh delay in seconds. Config: delay",
    show_default=True,
)
@click.option(
    "-e",
    "--exit",
    "_exit",
    type=click.Choice(["State", "Health"], case_sensitive=False),
    help="Set condition for watch exit.",
)
def osd(cluster, watch, delay, _exit):
    """
    Get basic state information on osd cluster
    """

    while watch:
        message = cli_cluster_state(cluster)
        click.clear()
        click.echo(message)
        state_exit_condition(_exit, message)
        sleep(delay)
    else:
        message = cli_cluster_state(cluster)
        click.echo(message)


@cli.command()
@click.argument("cluster")
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Are you sure you want to delete this cluster?",
    help="Confirm deletion on command execution.",
)
def delete(cluster):
    """
    Delete cluster.
    """
    cli_delete(cluster)


@cli.command()
def config():
    """
    Open configuration file
    """
    configuration_file()


@cli.command()
@click.option(
    "-c",
    "--cluster",
    default=try_function_defined(CONFIG, "addon", "cluster"),
    show_default=True,
    help="Name of cluster. Config: cluster",
)
@click.option(
    "-w/-x",
    "--watch/--no-watch",
    default=try_function_defined(CONFIG, "addon", "watch"),
    help="Watch the current state of a cluster. Config: watch [boolean]",
    show_default=True,
)
@click.option(
    "-d",
    "--delay",
    type=int,
    default=try_function_defined(CONFIG, "addon", "delay"),
    help="Refresh delay in seconds. Config: delay",
    show_default=True,
)
@click.option(
    "-p",
    "--prefix",
    type=str,
    help="Namespace prefix for addon. Config: prefix",
    default=try_function_defined(CONFIG, "addon", "prefix"),
    show_default=True,
)
def addon(cluster, watch, delay, prefix):
    """
    Get the current state of an installed addon instances
    """
    cli_addon(cluster, watch, delay, prefix)


@cli.command()
@click.option(
    "-c",
    "--config",
    "configuration",
    help="Configuration file used in build",
    default=try_function_defined(CONFIG, "index", "config"),
    show_default=True,
)
@click.option("--template", is_flag=True, help="Prints a sample configuration file")
def index(configuration, template):
    """
    Build and push images required for the doing catalog deployments.
    Build images for bundles, indexes and operator.
    This is done for the chain that can be created.
    """

    if template:
        cli_template()
    else:
        cli_index(configuration, CONFIG)


if __name__ == "__main__":
    cli()



build = {
    "configuration": {
        "remote": "git@github.com:integr8ly/integreatly-operator.git",
        "local": None,
        "tag_marker": "rhoam-v",
        "image_site": "quay.io",
        "org": None,
        "operator": "managed-api-service",
        "temporary": {"reuse": False, "location": None, "rebuild": False},
    },
    "chain": {
        "start": "latest",
        "service_affecting": "existing",
        "pre_release": None,
        "bundle": "build",
        "index": "build",
        "operator": "build",
    },
    "current": {},
    "latest": {"branch": "master", "service_affecting": False, "include": True},
    "new": {
        "branch": None,
        "location": "local",
        "include": True,
        "checkout": False,
        "ensure_branch": True,
        "service_affecting": True,
    },
}

required = {
    "configuration": ["local", "org"],
    "new": ["branch"],
}



import json
import pathlib
import secrets
import string
import subprocess  # nosec
import tempfile

import click
import requests


def create_kubeconfig():
    """
    creates a temporary kubeconfig in /tmp to be used for the script
    :return: str: path to file
    """

    suffix = []
    suffix_length = 7
    while suffix_length:
        suffix.append(secrets.choice(string.ascii_uppercase + string.digits))
        suffix_length -= 1

    suffix = "".join(suffix)
    kubeconfig = pathlib.Path(tempfile.gettempdir(), f"kudeconfig-{suffix}")

    with open(kubeconfig, "w"):
        if not kubeconfig.is_file():
            click.echo("Kubeconfig file not created")
            exit(1)
    return kubeconfig.absolute()


def get_addon_information(kube_config, route, bearer_token, name, prefix):
    cr = get_rhoam_cr(kube_config, prefix)
    if cr is None:
        click.secho("failed to get the rhmi cr", fg="red")
        exit(1)
    try:
        alert_data = get_addon_alerts(
            route, bearer_token
        )  # TODO alerts might not be there when start the script
    except requests.exceptions.ConnectionError:
        alert_data = None

    subprocess.call("clear")  # nosec
    rhoam_details(cr, name)
    click.echo()
    process_alerts(alert_data)


def login_to_cluster(creds, kube_config):
    data = subprocess.run(  # nosec
        [
            "oc",
            "login",
            "--kubeconfig",
            kube_config,
            "-u",
            creds["user"],
            "-p",
            creds["password"],
            "--server",
            creds["api"],
        ],
        capture_output=True,
    )

    if data.stderr != b"":
        click.echo(f"Error logging into cluster, {str(data.stderr)}")


def get_bearer_token(auth):
    """
    get the bearer token for the give user and cluster
    :param auth: dict: {}
    :return: str:
    """
    base = auth["api"]
    base: str = base.split(":")[1]
    base = base.replace("//api", "https://oauth-openshift.apps")

    url = f"{base}/oauth/authorize?client_id=openshift-challenging-client&response_type=token"
    s = requests.Session()
    r = s.get(
        url,
        headers={"X-CSRF-Token": "xxx"},
        cookies={},
        auth=(auth["user"], auth["password"]),
    )

    location = r.history[0].headers["Location"]
    location = location.split("#")[1]
    location = location.split("&")

    token = None
    for item in location:
        if item.startswith("access_token"):
            token = item.split("=")[1]

    return token


def get_prometheus_route(kube_config, prefix):
    cmd = ["oc", "--kubeconfig", kube_config, "get", "routes", "-A", "-o", "json"]
    data = subprocess.run(cmd, capture_output=True)  # nosec
    data = json.loads(data.stdout)

    for item in data["items"]:
        if item["metadata"]["name"] == "prometheus" and is_namespace(
            item["metadata"]["namespace"], prefix, "observability"
        ):
            return item["spec"]["host"]


def is_namespace(namespace, prefix, suffix, separator="-"):
    ns = f"{prefix}{separator}{suffix}"
    if namespace == ns:
        return True
    else:
        return False


def get_rhoam_cr(kube_config, prefix):
    """
    Get the RHOAM CR in json format
    :param kube_config:
    :param prefix
    :return:
    """
    cmd = [
        "oc",
        "--kubeconfig",
        kube_config,
        "get",
        "rhmi",
        "rhoam",
        "-n",
        f"{prefix}-operator",
        "-o",
        "json",
    ]
    data = subprocess.run(cmd, capture_output=True)  # nosec
    if data.stdout == b"":
        return None
    return json.loads(data.stdout)


def get_addon_alerts(route, bearer_token):
    resq = requests.get(
        f"https://{route}/api/v1/alerts",
        headers={"Authorization": f"Bearer {bearer_token}"},
        cookies={},
        auth=(),
    )

    if resq.status_code == 200:
        return resq.json()["data"]["alerts"]
    else:
        return None


def rhoam_details(CR, cluster):
    """
    Print details from rhoam CR
    :param CR:
    :param cluster
    :return:
    """

    click.echo(click.style("Cluster: ", bold=True) + click.style(cluster))

    if "status" not in CR:
        click.secho("No status block in CR yet", fg="red")
        return

    status = CR["status"]["stage"]
    click.echo(click.style("Status: ", bold=True) + click.style(status))
    if "version" in CR["status"]:
        click.echo(
            click.style("Version: ", bold=True) + click.style(CR["status"]["version"])
        )
    if "toVersion" in CR["status"]:
        click.echo(
            click.style("To Version: ", bold=True)
            + click.style(CR["status"]["toVersion"])
        )
    if "quota" in CR["status"]:
        click.echo(
            click.style("Quota: ", bold=True) + click.style(CR["status"]["quota"])
        )
    if "toQuota" in CR["status"]:
        click.echo(
            click.style("To Quota: ", bold=True) + click.style(CR["status"]["toQuota"])
        )
    if "lastError" in CR["status"] and len(CR["status"]["lastError"]) > 0:
        start = click.style(f"Last Error:\n{'-' * 20}\n", bold=True)
        middle = click.style(CR["status"]["lastError"])
        end = click.style(f"\n{'-' * 20}")
        click.echo(f"{start}{middle}{end}")
    if status != "complete":
        click.echo(click.style("Current stage is: ", bold=True) + click.style(status))

        for stage in CR["status"]["stages"]:
            if stage == "bootstrap":
                continue
            for product in CR["status"]["stages"][stage]["products"]:
                if (
                    CR["status"]["stages"][stage]["products"][product]["status"]
                    != "completed"
                ):
                    current_product = CR["status"]["stages"][stage]["products"][product]

                    click.echo(
                        f"\t{current_product['name']}, {current_product['status']}"
                    )


def process_alerts(data: dict):
    if data is None:
        click.echo("No alert data found")
        return

    alerts = {}

    for point in data:
        if point["state"] in alerts:
            alerts[point["state"]].append(point)
        else:
            alerts.setdefault(point["state"], [])
            alerts[point["state"]].append(point)

    # TODO loops need to be sorted. They keep flipping
    for status in alerts:
        click.echo(
            click.style(f"{status} alerts: ", bold=True)
            + click.style(f"{len(alerts[status])}", fg="red")
        )

    click.echo()
    for status in alerts:
        click.echo(click.style(f"{status} alerts: ", bold=True))
        for item in alerts[status]:
            click.echo(f"\t{item['labels']['alertname']}")
        click.echo()



import json
import os
import subprocess  # nosec
import sys
from time import sleep

import click
import pyperclip

from orc_dev_cli.addon_state import (
    create_kubeconfig,
    get_addon_information,
    get_bearer_token,
    get_prometheus_route,
    login_to_cluster,
)


def cli_creds(cluster):
    console, creds = get_cluster_credentials(cluster)
    pyperclip.copy(creds["password"])
    print_screen(console, creds, cluster)


def cli_delete(cluster):
    click.echo("Deleting Cluster: " + click.style(cluster, fg="red"))
    cluster = get_cluster_id(cluster)
    delete_cluster(cluster)


def get_cluster_id(cluster_name):
    found_cluster = get_cluster_data(cluster_name)

    cluster_id: str = found_cluster["id"]
    return cluster_id


def get_cluster_data(cluster_name):
    output = subprocess.run(["ocm", "get", "clusters"], capture_output=True)  # nosec
    data = json.loads(output.stdout)
    data = data["items"]
    found_cluster = None
    for item in data:
        if item["name"] == cluster_name:
            found_cluster = item
            break
    if found_cluster is None:
        click.echo(f'No cluster with display name "{cluster_name}" was found')
        exit(1)
    return found_cluster


def delete_cluster(cluster_id):
    subprocess.run(["ocm", "delete", "cluster", cluster_id])  # nosec


def print_screen(console_url, data, name):
    click.echo(click.style("Cluster: ", bold=True) + click.style(name))
    click.echo(click.style("Console: ", bold=True) + click.style(console_url))
    click.echo(
        click.style("Console Login: ", bold=True)
        + click.style(get_console_login(base_url(console_url)))
    )

    click.echo()
    click.echo(click.style("Login Command:", bold=True))
    click.echo(login_command(data))

    click.echo()
    click.echo(click.style("User: ", bold=True) + click.style(data["user"]))
    click.echo(
        click.style("Password: ", bold=True)
        + click.style(data["password"])
        + click.style(" (copied)", italic=True, dim=True)
    )


def base_url(url):
    spilt = url.split("apps.")
    return spilt[1]


def login_command(creds):
    cmd = [
        "oc",
        "login",
        "-u",
        creds["user"],
        "-p",
        creds["password"],
        "--server",
        creds["api"],
    ]
    return " ".join(cmd)


def get_console_login(url):
    return (
        f"https://oauth-openshift.apps.{url}/"
        f"login?then=%2Foauth%2Fauthorize%3Fclient_id%3Dconsole%26idp%3Dkubeadmin%26"
        f"redirect_uri%3Dhttps%253A%252F%252Fconsole-openshift-console.apps."
        f"{url}%252Fauth%252Fcallback%26response_type%3Dcode%26scope%3Duser%253Afull"
    )


def get_cluster_credentials(cluster_name):
    """
    Get the login details for the cluster admin
    :param cluster_name:
    :return: console url, {'user': <user>, 'password': <password>, 'api': <api>}
    """

    found_cluster = get_cluster_data(cluster_name)
    cluster_id: str = found_cluster["id"]
    cluster_console = found_cluster["console"]["url"]
    api = found_cluster["api"]["url"]
    cmd = f"ocm get /api/clusters_mgmt/v1/clusters/{cluster_id}/credentials"
    data = os.popen(  # nosec
        cmd
    )  # don't know why I can't get this to run in subprocess.run
    data = json.loads(data.read())["admin"]
    data["api"] = api
    return cluster_console, data


def get_cluster_state(cluster_data):
    state = cluster_data["status"]["state"]
    return state


def get_cluster_health(cluster_data):
    try:
        health = cluster_data["items"][0]["metrics"][0]["health_state"]
    except KeyError:
        health = "Unknown"
    return health


def get_cluster_sub(cluster):
    cmd = subprocess.run(  # nosec
        ["ocm", "get", "subs", "--parameter", f"search=cluster_id='{cluster}'"],
        capture_output=True,
    )
    if cmd.stderr is None:
        click.secho("Error getting subs form cluster", fg="red")
        sys.exit(1)
    data = json.loads(cmd.stdout)
    return data


def cli_cluster_state(name):
    cluster_id = get_cluster_id(name)
    data1 = get_cluster_data(name)
    data = get_cluster_sub(cluster_id)

    state = get_cluster_state(data1)
    out = f"Cluster: {name}\nState: "
    if state == "ready":
        state = state.upper()
        out += click.style(state, fg="green")
    else:
        out += state
        return out

    out += "\nHealth: "
    health = get_cluster_health(data)
    if health == "healthy":
        health = health.upper()
        out += click.style(health, fg="green")
    else:
        out += health

    return out


def state_exit_condition(state, message):

    if state is None:
        return

    state = state.lower()
    data = message.split("\n")

    if len(data) > 2:
        status = {"state": data[1], "health": data[2]}
    else:
        status = {"state": data[1], "health": ""}

    if state == "state" and "READY" in status[state]:
        sys.exit()

    if state == "health" and "HEALTHY" in status[state]:
        sys.exit()

    return


def cli_create_cluster(name, version, region, node, num):
    click.echo("create a cluster")
    click.echo(f"name: {name}")
    click.echo(f"version: {version}")
    click.echo(f"region: {region}")
    click.echo(f"node: {node}")
    click.echo(f"num: {num}")


def cli_addon(name, watch, delay, prefix):
    kubeconfig = create_kubeconfig()
    _, creds = get_cluster_credentials(name)
    login_to_cluster(creds, kubeconfig)
    bearer_token = get_bearer_token(creds)
    route = get_prometheus_route(kubeconfig, prefix)

    while watch:
        get_addon_information(kubeconfig, route, bearer_token, name, prefix)
        sleep(delay)
        if route is None:
            route = get_prometheus_route(kubeconfig, prefix)
    else:
        get_addon_information(kubeconfig, route, bearer_token, name, prefix)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['orc_dev_cli', 'orc_dev_cli.index_build']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.28,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'click>=8.1.2,<9.0.0',
 'podman>=4.2.0,<5.0.0',
 'pyperclip>=1.8.2,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'semver>=2.13.0,<3.0.0']

extras_require = \
{':python_version < "3.11"': ['toml>=0.10.2,<0.11.0']}

entry_points = \
{'console_scripts': ['orc = orc_dev_cli.cli:cli']}

setup_kwargs = {
    'name': 'orc-dev-cli',
    'version': '0.6.1',
    'description': 'CLI dev tool for interacting with openshift clusters and RHOAM addons.',
    'long_description': '# Welcome to orc-dev-cli\n\nOrc-dev-cli gives developers a CLI tool called orc.\nOrc is used to interact with openshift osd clusters and the openshift addon RHOAM.\n\nAs this is a dev tool kude-admin access to the clusters is expected.\n\n## Installation\nEasiest way to install orc is by using pip\n```shell\npip install orc-dev-cli\n```\n\n## Basic Commands\n\n* `orc addon`   Get the current state of an installed addon instances\n* `orc config`  Open configuration file\n* `orc delete`  Delete cluster.\n* `orc login`   Get cluster kubeadmin login details.\n* `orc osd`     Get basic state information on osd cluster\n\n## Shell Completion\n\n### Bash\n\n1. Create temporary bash script.\n2. Add `_ORC_COMPLETE=bash_source orc > ~/.orc-complete.bash`.\n3. Run script.\n4. Source the newly created `~/.orc-complete.bash` in `~/.bashrc` as follows `. ~/.orc-complete.bash`.\n5. Restart the shell to complete.\n',
    'author': 'Jim Fitzpatrick',
    'author_email': 'jimfity@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Boomatang/orc-dev-cli',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import json
import logging
import os
import shutil
import subprocess  # nosec
import sys
import tempfile
from pathlib import Path
from pprint import pprint
from typing import Dict

import click
import git
import requests
import semver
import yaml
from podman import PodmanClient

from orc_dev_cli import config
from orc_dev_cli.config import merge
from orc_dev_cli.index_build.default import build, required

if sys.version_info.minor < 11:
    import toml
else:
    import tomllib as toml


def cli_template():
    with open("samples/index_build.toml") as f:
        click.echo(f.read())


def valid_config(config: dict, must_have: dict) -> (bool, list):
    # TODO this function needs to be filled out
    return True, []


def clone_repo(repo: str, location: Path):
    if not repo.startswith("git@") or not repo.endswith(".git"):
        click.secho(
            f"Repo clone url seems to be in correct please check configuration. {repo}",
            fg="red",
            bold=True,
        )
        exit(1)

    click.echo(f"Cloning repo: {repo} to {location}")
    r = git.Repo.clone_from(repo, location)
    return r


def existing_repo(location: Path):
    return git.Repo(location)


def get_release(repo: git.Repo, prefix):
    tags = repo.tags
    releases = (t for t in tags if t.name.lower().startswith(prefix))
    versions = {}

    for release in releases:
        version = get_semver(release, prefix)
        versions[version] = release

    return versions


def sort_releases(releases):

    keys = list(releases.keys())
    keys = sorted(keys)

    versions = []
    prerelease = []
    prerelease_rc = {}
    for key in keys:
        if not key.prerelease:
            versions.append(key)
        else:
            prerelease.append(key)

    for release in prerelease:
        a = semver.VersionInfo(
            major=release.major, minor=release.minor, patch=release.patch
        )
        if a not in versions:
            if a in prerelease_rc.keys():
                prerelease_rc[a].append(release)
            else:
                prerelease_rc.setdefault(
                    a,
                    [
                        release,
                    ],
                )

    for rcs in prerelease_rc:
        high = None
        for r in prerelease_rc[rcs]:
            if high is None:
                high = r
            else:
                if high.prerelease < r.prerelease:
                    high = r
        versions.append(high)

    versions = sorted(versions)
    return versions


def none_semver(config):
    """
    This function will return the latest and new markers if the version are in the config
    """
    result = []
    other = ["latest", "new"]

    if config is None:
        return result

    for key in other:
        if key in config and config[key]["include"]:
            result.append(key)

    return result


def get_working_release(chain_start, data):

    match chain_start:
        case None:
            return {"tags": [], "other": []}
        case "new":
            return {"tags": [], "other": ["new"]}
        case "latest":
            return {"tags": [], "other": ["latest", "new"]}
        case "current":
            return {"tags": [data["tags"][-1]], "other": data["other"]}
        case _:
            base = semver.VersionInfo.parse(chain_start)
            index = data["tags"].index(base)
            tags = data["tags"][index:]
            return {"tags": tags, "other": data["other"]}


def checkout(repo, tag, config):
    if type(tag) == semver.VersionInfo:
        name = f"{config['configuration']['tag_marker']}{tag}"
    else:
        name = tag

    repo.git.checkout(name)


def get_service_affecting(csv):
    with open(csv, "r") as f:
        data = yaml.safe_load(f)

        if (
            "metadata" in data
            and "annotations" in data["metadata"]
            and "serviceAffecting" in data["metadata"]["annotations"]
        ):
            return data["metadata"]["annotations"]["serviceAffecting"]
    return True


def remove_replaces(csv):
    with open(csv, "r") as f:
        data = yaml.safe_load(f)

    del data["spec"]["replaces"]
    with open(csv, "w") as f:
        f.write(yaml.dump(data))


def container_exist_remote(config, repo, tag: str):
    image_site = config["configuration"]["image_site"]
    org = config["configuration"]["org"]

    if image_site != "quay.io":
        click.secho(
            "The checking for existing containers only works with quay.io",
            fg="red",
            bold=True,
        )
        return False

    click.echo(f"Searching for tag {tag} in repository {repo}")
    url = f"https://{image_site}/api/v1/repository/{org}/{repo}/tag"
    resq = requests.get(url)
    data = resq.json()
    data = data["tags"]
    for t in data:
        if t["name"] == tag:
            return True
    return False


def build_container(image, uri, container_file, dockerfile=None, buildargs=None):
    click.echo(f"Building container image: {image}")
    click.secho(f"Container File: {container_file}", dim=True)
    with PodmanClient(base_url=uri) as client:
        if dockerfile is None:
            client.images.build(
                path=container_file.parent,
                dockerfile=container_file.name,
                tag=image,
                buildargs=buildargs,
            )
        else:
            click.secho(f"Building with dockerfile: {dockerfile}", dim=True)
            client.images.build(
                path=container_file,
                dockerfile=dockerfile,
                tag=image,
                buildargs=buildargs,
            )
        client.images.prune()


def push_container(image, uri):
    click.echo(f"Pushing container image: {image}")
    with PodmanClient(base_url=uri) as client:
        client.images.push(image)


def build_index(image, bundles):
    click.echo(f"Building index image: {image}")
    click.secho(f"Bundles in index: {bundles}", dim=True)
    b = []
    for bundle in bundles:
        b.append("--bundles")
        b.append(bundle)

    s = ["opm", "index", "add"]
    e = ["--build-tool", "podman", "--tag", image]
    cmd = s + b + e
    subprocess.run(cmd, check=True)  # nosec


def work_on_tag(repo, tag, config, bundles, first=False, label=None):
    uri = f"unix:/run/user/{os.getuid()}/podman/podman.sock"
    csv = Path(
        repo.working_dir,
        "bundles",
        config["configuration"]["operator"],
        f"{tag.major}.{tag.minor}.{tag.patch}",
        "manifests",
        f"{config['configuration']['operator']}.clusterserviceversion.yaml",
    )

    image_site = config["configuration"]["image_site"]
    org = config["configuration"]["org"]
    operator = config["configuration"]["operator"]
    bundle = f"{operator}-bundle"
    index = f"{operator}-index"
    config_tag = semver.VersionInfo(major=tag.major, minor=tag.minor, patch=tag.patch)

    pre_release = get_config_value("pre_release", config, config_tag, label)

    if pre_release not in [False, None]:
        tag = semver.VersionInfo(
            major=tag.major,
            minor=tag.minor,
            patch=tag.patch,
            prerelease=pre_release,
        )
    else:
        tag = semver.VersionInfo(major=tag.major, minor=tag.minor, patch=tag.patch)
    operator_tag = f'{config["configuration"]["tag_marker"]}{tag}'
    data = {
        "label": label,
        "operator": f"{Path(image_site, org, operator)}:{operator_tag}",
        "bundle": f"{Path(image_site, org, bundle)}:{tag}",
        "index": f"{Path(image_site, org, index)}:{tag}",
    }
    bundles.append(data["bundle"])

    build_operator = True
    if get_config_value("operator", config, config_tag, label) == "reuse":
        result = container_exist_remote(config, operator, str(operator_tag))
        if result:
            click.secho(
                f"Reusing existing remote operator: {data['operator']}", fg="green"
            )
            build_operator = False

    do_build_index = True
    if get_config_value("index", config, config_tag, label) == "reuse":
        result = container_exist_remote(config, index, str(tag))
        if result:
            click.secho(f"Reusing existing remote index: {data['index']}", fg="green")
            do_build_index = False

    build_bundle = True
    if get_config_value("bundle", config, config_tag, label) == "reuse":
        result = container_exist_remote(config, bundle, str(tag))
        if result:
            click.secho(f"Reusing existing remote bundle: {data['bundle']}", fg="green")
            data["service_affecting"] = "unknown"
            build_bundle = False

    if build_bundle:
        state = get_config_value("service_affecting", config, config_tag, label)
        service_affecting = is_service_affecting(csv, state)
        data["service_affecting"] = service_affecting
        release_prepare(config, repo, service_affecting, tag)

        if first:
            remove_replaces(csv)

        data["bundle_built"] = True
        config["configuration"]["temporary"]["rebuild"] = True

    data["bundles"] = bundles.copy()

    if build_operator:
        build_container(data["operator"], uri, Path(repo.working_dir, "Dockerfile"))
        push_container(data["operator"], uri)

    if build_bundle:
        build_container(
            data["bundle"],
            uri,
            repo.working_dir,
            dockerfile=Path(
                "bundles", config["configuration"]["operator"], "bundle.Dockerfile"
            ),
            buildargs={"version": f"{tag.major}.{tag.minor}.{tag.patch}"},
        )
        push_container(data["bundle"], uri)

    if do_build_index or config["configuration"]["temporary"]["rebuild"]:
        click.secho("No validation is happening on the bundles", fg="red")
        build_index(data["index"], bundles)
        push_container(data["index"], uri)

    if label != "new":
        repo.git.restore("*")

    return data


def is_service_affecting(csv, state):

    match state:
        case "existing":
            if csv.exists():
                service_affecting = get_service_affecting(csv)
            else:
                service_affecting = True
            return service_affecting
        case _:
            return state


def get_config_value(setting, config, tag=None, label=None):

    if tag is not None:
        tag = str(tag)
        if tag in config and setting in config[tag]:
            return config[tag][setting]

    if label is not None:
        if label in config and setting in config[label]:
            return config[label][setting]

    state = config["chain"][setting]
    return state


def release_prepare(config, repo, service_affecting, tag):
    subprocess.run(  # nosec
        [
            "make",
            "release/prepare",
            f"SEMVER={tag}",
            f"ORG={config['configuration']['org']}",
            f"OLM_TYPE={config['configuration']['operator']}",
            f"SERVICE_AFFECTING={service_affecting}",
        ],
        cwd=repo.working_dir,
        check=True,
    )


def next_version(tag: semver.VersionInfo):
    return tag.bump_minor()


def write_json_config(chain):
    _chain = {}
    for key in chain:
        _chain[str(key)] = chain[key]

    data = json.dumps(_chain)
    data_file = Path(tempfile.gettempdir(), "orc_index_data.json")
    with open(data_file, "w") as f:
        f.write(data)


def cli_index(configuration, main_configuration=None):
    if main_configuration is None:
        main_configuration = {}
    configuration = Path(configuration)
    config = load_config(configuration, main_configuration)
    ok, errors = valid_config(config, required)
    if not ok:
        for error in errors:
            click.secho(error, fg="red")
        exit(1)
    temp_location = Path(tempfile.gettempdir(), config["configuration"]["operator"])
    config["configuration"]["temporary"]["location"] = temp_location

    if config["configuration"]["temporary"]["reuse"] and temp_location.exists():
        click.echo("Using existing temporary repo")
        repo = existing_repo(temp_location)
    else:
        if temp_location.exists():
            click.echo("Removing existing temporary repo")
            shutil.rmtree(temp_location)
        repo = clone_repo(config["configuration"]["remote"], temp_location)

    releases = get_release(repo, config["configuration"]["tag_marker"])
    versions = sort_releases(releases)
    other = none_semver(config)
    data = {"tags": versions, "other": other}
    data = get_working_release(config["chain"]["start"], data)

    chain = {}
    bundles = []
    last_tag = None
    for tag in data["tags"]:
        click.echo(f"Working on: {tag}")
        if data["tags"].index(tag) == 0:
            first = True
        else:
            first = False
        checkout(repo, tag, config)

        label = None
        if tag == data["tags"][-1]:
            label = "current"
        chain_data = work_on_tag(repo, tag, config, bundles, first, label=label)
        chain[tag] = chain_data
        last_tag = tag

    if "latest" in data["other"]:

        click.echo(f"Working on: latest ({config['latest']['branch']})")

        if last_tag is None:
            first = True
        else:
            first = False

        branch = config["latest"]["branch"]
        checkout(repo, branch, config)

        bundle_path = Path(
            repo.working_dir, "bundles", config["configuration"]["operator"]
        )

        if config["chain"]["start"] == "latest":
            tag = get_version_latest_tag(bundle_path)
        else:
            tag = next_version(last_tag)

        chain_data = work_on_tag(repo, tag, config, bundles, first, label="latest")
        chain[tag] = chain_data
        last_tag = tag

    if "new" in data["other"]:
        chain_data, tag = build_new(bundles, config, last_tag)
        chain[tag] = chain_data

    write_json_config(chain)


def get_version_latest_tag(bundle_path):
    versions = []
    dirs = os.listdir(bundle_path)
    for parts in dirs:
        if Path(bundle_path, str(parts)).is_dir():
            versions.append(semver.VersionInfo.parse(parts))
    versions = sorted(versions)
    return versions[-1]


def build_new(bundles, _config, last_tag):
    click.echo(f"Working on: new ({_config['new']['branch']})")
    if last_tag is None:
        first = True
    else:
        first = False
    branch = _config["new"]["branch"]

    location = _config["new"]["location"]
    if location == "local":
        repo = existing_repo(_config["configuration"]["local"])
    elif location == "temp":
        repo = existing_repo(_config["configuration"]["temporary"]["location"])
    else:
        repo = None
        click.secho("Invalid location set on 'new'", fg="red")
        exit(1)

    ok = _config["new"]["checkout"]
    if ok:
        checkout(repo, branch, _config)

    ensure_branch = _config["new"]["ensure_branch"]
    if ensure_branch:
        if repo.head.is_detached:
            click.secho(
                "Current branch does not match required branch: Repo head in detached state",
                fg="red",
            )
            exit(1)
        if repo.active_branch.name != branch:
            click.secho(
                f"Current branch does not match required branch: {branch} != {repo.active_branch.name}",
                fg="red",
            )
            exit(1)

    bundle_path = Path(
        repo.working_dir, "bundles", _config["configuration"]["operator"]
    )
    if _config["chain"]["start"] == "new":
        tag = get_version_latest_tag(bundle_path)
    else:
        tag = next_version(last_tag)
        csv = Path(
            bundle_path,
            f"{tag.major}.{tag.minor}.{tag.patch}",
            "manifests",
            f"{_config['configuration']['operator']}.clusterserviceversion.yaml",
        )
        if not csv.exists():
            click.echo(f"Build {last_tag} to ensure the chain")
            release_prepare(_config, repo, True, last_tag)
    chain_data = work_on_tag(repo, tag, _config, bundles, first, label="new")
    return chain_data, tag


def get_semver(tag: git.TagReference, prefix=None):
    if prefix is None:
        name = tag.name
    else:
        name = tag.name.strip(prefix)

    return semver.VersionInfo.parse(str(name))


def load_config(config_file: Path, main_config_file: Dict):
    data = {}
    _build = build

    if "index" in main_config_file:
        if "configuration" in main_config_file["index"]:
            _build = merge(
                _build, {"configuration": main_config_file["index"]["configuration"]}
            )

        if "chain" in main_config_file["index"]:
            _build = merge(_build, {"chain": main_config_file["index"]["chain"]})

    if config_file.is_file():
        with open(config_file, "r") as f:
            data = toml.loads(f.read())

    output = merge(_build, data)

    return output


if "__main__" == __name__:
    cli_index(
        "/home/jimfitz/code/github.com/Boomatang/orc-dev-cli/samples/simple.toml",
        config.load_config(),
    )



import pathlib
import subprocess  # nosec
import sys
import textwrap

import click

if sys.version_info.minor < 11:
    import toml
else:
    import tomllib as toml

HOST = pathlib.Path.home()
CONFIG_PATH = pathlib.Path(HOST, ".config", "orc")
CONFIG_FILE = pathlib.Path(CONFIG_PATH, "config.toml")


def configuration_file():

    if not CONFIG_FILE.is_file():
        CONFIG_PATH.mkdir(parents=True, exist_ok=True)
        data = textwrap.dedent(
            """\
        [default]
        cluster = ""
        delay = 10
        prefix = "redhat-rhoam"
        watch = false
        """
        )
        with open(CONFIG_FILE, "w") as f:
            f.write(data)

        click.echo(f"Configuration file created: {CONFIG_FILE}")

    cmd = None
    if sys.platform.startswith("linux"):
        cmd = "xdg-open"
    elif sys.platform.startswith("darwin"):
        cmd = "open"
    else:
        click.echo("Auto opening of configuration file failed unsupported platform")
        exit(1)

    subprocess.call([cmd, CONFIG_FILE])  # nosec


def merge(a: dict, b: dict):

    for key in a:
        if key in b:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                a[key].update(merge(a[key], b[key]))
            else:
                a[key] = b[key]

    for key in b:
        if key not in a:
            a[key] = b[key]

    return a


def load_config():
    data = {}
    config = {
        "default": {
            "cluster": "",
            "delay": 10,
            "prefix": "redhat-rhoam",
            "watch": False,
        },
    }

    if CONFIG_FILE.is_file():
        with open(CONFIG_FILE, "r") as f:
            data = toml.loads(f.read())

    output = merge(config, data)

    return output


def try_function_defined(config: dict, func: str, field: str):
    if func in config:
        if field in config[func]:
            return config[func][field]
    try:
        return config["default"][field]
    except KeyError:
        return None
